#!/usr/bin/env python

"""
DOM wrapper around libxml2, specifically the libxml2mod Python extension module.
"""

__version__ = "0.2"

import libxml2
from libxml2dom.macrolib import *
from libxml2dom.macrolib import \
    createDocument as Node_createDocument, \
    parseString as Node_parseString, parseURI as Node_parseURI, \
    parseFile as Node_parseFile, \
    toString as Node_toString, toStream as Node_toStream, \
    toFile as Node_toFile
import weakref

# Attribute and node list wrappers.

class NamedNodeMap(object):

    """
    A wrapper around Node objects providing DOM and dictionary convenience
    methods.
    """

    def __init__(self, node):
        self.node = node

    def getNamedItem(self, name):
        return self.node.getAttributeNode(name)

    def getNamedItemNS(self, ns, localName):
        return self.node.getAttributeNodeNS(ns, localName)

    def setNamedItem(self, node):
        self.node.setAttributeNode(node)
        # NOTE: Non-standard return of attribute node, since the actual node is
        # NOTE: not set on the target node.
        return self.getNamedItem(node.name)

    def setNamedItemNS(self, node):
        self.node.setAttributeNodeNS(node)
        # NOTE: Non-standard return of attribute node, since the actual node is
        # NOTE: not set on the target node.
        return self.getNamedItemNS(node.namespaceURI, node.name)

    # Dictionary emulation methods.

    def __getitem__(self, name):
        return self.getNamedItem(name)

    def __setitem__(self, name, node):
        if name == node.nodeName:
            self.setNamedItem(node)
        else:
            raise KeyError, name

    def __delitem__(self, name):
        # NOTE: To be implemented.
        pass

    def values(self):
        return [Attribute(_node, self.node.ownerDocument) for _node in Node_attributes(self.node.as_native_node()).values()]

    def keys(self):
        return [(attr.namespaceURI, attr.localName) for attr in self.values()]

    def items(self):
        return [((attr.namespaceURI, attr.localName), attr) for attr in self.values()]

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "{%s}" % ",\n".join(["%s : %s" % (repr(key), repr(value)) for key, value in self.items()])

class NodeList(list):

    "A wrapper around node lists."

    def item(self, index):
        return self[index]

    def length(self):
        return len(self)

# Node classes.

class Node(object):

    """
    A DOM-style wrapper around libxml2mod objects.
    """

    ATTRIBUTE_NODE = xml.dom.Node.ATTRIBUTE_NODE
    COMMENT_NODE = xml.dom.Node.COMMENT_NODE
    DOCUMENT_NODE = xml.dom.Node.DOCUMENT_NODE
    DOCUMENT_TYPE_NODE = xml.dom.Node.DOCUMENT_TYPE_NODE
    ELEMENT_NODE = xml.dom.Node.ELEMENT_NODE
    ENTITY_NODE = xml.dom.Node.ENTITY_NODE
    ENTITY_REFERENCE_NODE = xml.dom.Node.ENTITY_REFERENCE_NODE
    NOTATION_NODE = xml.dom.Node.NOTATION_NODE
    PROCESSING_INSTRUCTION_NODE = xml.dom.Node.PROCESSING_INSTRUCTION_NODE
    TEXT_NODE = xml.dom.Node.TEXT_NODE

    def __init__(self, node, ownerDocument=None):
        self._node = node
        self.ownerDocument = ownerDocument

    def as_native_node(self):
        return self._node

    def _nodeType(self):
        return Node_nodeType(self._node)

    def _childNodes(self):

        # NOTE: Consider a generator instead.

        return NodeList([Node(_node, self.ownerDocument) for _node in Node_childNodes(self._node)])

    def _attributes(self):
        return NamedNodeMap(self)

    def _namespaceURI(self):
        return Node_namespaceURI(self._node)

    def _nodeValue(self):
        return Node_nodeValue(self._node)

    def _setNodeValue(self, value):
        Node_setNodeValue(self._node, value)

    def _prefix(self):
        return Node_prefix(self._node)

    def _nodeName(self):
        return Node_nodeName(self._node)

    def _tagName(self):
        return Node_tagName(self._node)

    def _localName(self):
        return Node_localName(self._node)

    def _parentNode(self):
        return get_node(Node_parentNode(self._node), self)

    def _previousSibling(self):
        return Node(Node_previousSibling(self._node), self.ownerDocument)

    def _nextSibling(self):
        return Node(Node_nextSibling(self._node), self.ownerDocument)

    def hasAttributeNS(self, ns, localName):
        return Node_hasAttributeNS(self._node, ns, localName)

    def hasAttribute(self, name):
        return Node_hasAttribute(self._node, name)

    def getAttributeNS(self, ns, localName):
        return Node_getAttributeNS(self._node, ns, localName)

    def getAttribute(self, name):
        return Node_getAttribute(self._node, name)

    def getAttributeNodeNS(self, ns, localName):
        return Attribute(Node_getAttributeNodeNS(self._node, ns, localName), self.ownerDocument)

    def getAttributeNode(self, localName):
        return Attribute(Node_getAttributeNode(self._node, localName), self.ownerDocument)

    def setAttributeNS(self, ns, name, value):
        Node_setAttributeNS(self._node, ns, name, value)

    def setAttribute(self, name, value):
        Node_setAttribute(self._node, name, value)

    def setAttributeNodeNS(self, node):
        Node_setAttributeNodeNS(self._node, node._node)

    def setAttributeNode(self, node):
        Node_setAttributeNode(self._node, node._node)

    def createElementNS(self, ns, name):
        return Node(Node_createElementNS(self._node, ns, name), self.ownerDocument)

    def createElement(self, name):
        return Node(Node_createElement(self._node, name), self.ownerDocument)

    def createAttributeNS(self, ns, name):
        tmp = self.createElement("tmp")
        return Attribute(Node_createAttributeNS(tmp._node, ns, name))

    def createAttribute(self, name):
        tmp = self.createElement("tmp")
        return Attribute(Node_createAttribute(tmp._node, name))

    def createTextNode(self, value):
        return Node(Node_createTextNode(self._node, value), self.ownerDocument)

    def createComment(self, value):
        return Node(Node_createComment(self._node, value), self.ownerDocument)

    def importNode(self, node, deep):
        if hasattr(node, "as_native_node"):
            return Node(Node_importNode(self._node, node.as_native_node(), deep), self.ownerDocument)
        else:
            return Node(Node_importNode_DOM(self._node, node, deep), self.ownerDocument)

    def insertBefore(self, tmp, oldNode):
        if hasattr(tmp, "as_native_node"):
            return Node(Node_insertBefore(self._node, tmp.as_native_node(), oldNode.as_native_node()), self.ownerDocument)
        else:
            return Node(Node_insertBefore(self._node, tmp, oldNode.as_native_node()), self.ownerDocument)

    def replaceChild(self, tmp, oldNode):
        if hasattr(tmp, "as_native_node"):
            return Node(Node_replaceChild(self._node, tmp.as_native_node(), oldNode.as_native_node()), self.ownerDocument)
        else:
            return Node(Node_replaceChild(self._node, tmp, oldNode.as_native_node()), self.ownerDocument)

    def appendChild(self, tmp):
        if hasattr(tmp, "as_native_node"):
            return Node(Node_appendChild(self._node, tmp.as_native_node()), self.ownerDocument)
        else:
            return Node(Node_appendChild(self._node, tmp), self.ownerDocument)

    def removeChild(self, tmp):
        if hasattr(tmp, "as_native_node"):
            Node_removeChild(self._node, tmp.as_native_node())
        else:
            Node_removeChild(self._node, tmp)

    childNodes = property(_childNodes)
    value = data = nodeValue = property(_nodeValue, _setNodeValue)
    name = nodeName = property(_nodeName)
    tagName = property(_tagName)
    namespaceURI = property(_namespaceURI)
    prefix = property(_prefix)
    localName = property(_localName)
    parentNode = property(_parentNode)
    nodeType = property(_nodeType)
    attributes = property(_attributes)
    previousSibling = property(_previousSibling)
    nextSibling = property(_nextSibling)

    #def isSameNode(self, other):
    #    return self._node.nodePath() == other._node.nodePath()

    #def __eq__(self, other):
    #    return self._node.nodePath() == other._node.nodePath()

    def getElementsByTagName(self, tagName):
        return self.xpath("//" + tagName)

    # 4DOM extensions to the usual PyXML API.
    # NOTE: To be finished.

    def xpath(self, expr, variables=None, namespaces=None):
        result = Node_xpath(self._node, expr, variables, namespaces)
        if hasattr(result, "__len__"):
            return NodeList([get_node(_node, self) for _node in result])
        else:
            return result

    # Convenience methods.

    def toString(self, encoding=None):
        return toString(self, encoding)

    def toStream(self, stream, encoding=None):
        toStream(self, stream, encoding)

    def toFile(self, f, encoding=None):
        toFile(self, f, encoding)

# Attribute nodes.

class Attribute(Node):

    "A class providing attribute access."

    def __init__(self, node, ownerDocument=None, ownerElement=None):
        Node.__init__(self, node, ownerDocument)
        self.ownerElement = ownerElement

    def _parentNode(self):
        return self.ownerElement

    parentNode = property(_parentNode)

# Document housekeeping mechanisms.

class Document(Node):

    "A class providing document-level housekeeping."

    def __init__(self, node):
        self._node = node
        self.weakref_ownerDocument = weakref.ref(self)

    def _ownerDocument(self):
        return self.weakref_ownerDocument()

    def __del__(self):
        #print "Freeing document", self._node
        libxml2mod.xmlFreeDoc(self._node)

    ownerDocument = property(_ownerDocument)

# Factory functions.

def get_node(_node, context_node):
    if Node_nodeType(_node) == context_node.DOCUMENT_NODE:
        return context_node.ownerDocument
    elif Node_nodeType(_node) == context_node.ATTRIBUTE_NODE:
        return Attribute(_node, context_node.ownerDocument, context_node)
    else:
        return Node(_node, context_node.ownerDocument)

# Utility functions.

def createDocumentType(localName, publicId, systemId):
    return None

def createDocument(namespaceURI, localName, doctype):
    return Document(Node_createDocument(namespaceURI, localName, doctype))

def parse(stream_or_string):
    if hasattr(stream_or_string, "read"):
        stream = stream_or_string
        return parseString(stream.read())
    else:
        return parseFile(stream_or_string)

def parseFile(s):
    return Document(Node_parseFile(s))

def parseString(s):
    return Document(Node_parseString(s))

def parseURI(uri):
    return Document(Node_parseURI(uri))

def toString(node, encoding=None):
    return Node_toString(node.as_native_node(), encoding)

def toStream(node, stream, encoding=None):
    Node_toStream(node.as_native_node(), stream, encoding)

def toFile(node, f, encoding=None):
    Node_toFile(node.as_native_node(), f, encoding)

def adoptNodes(nodes):
    if len(nodes) == 0:
        return []
    doc = Document(libxml2mod.doc(nodes[0]))
    results = []
    for node in nodes:
        results.append(Node(node, doc))
    return results

# vim: tabstop=4 expandtab shiftwidth=4
